// Piotr Daski	E1ISIII

// W tym pliku nagwkowym jest zgromadzona caa biblioteka Variant
#include <boost/variant.hpp>

// Sama biblioteka zostaa podzielona na czci, ktre mona docza wybirczo zamiast caoci
//      "boost/varian/variant_fwd.hpp"   - deklaracje zapowiadajce szablonw klasy variant
//      "boost/varian/variant.hpp"       - definicje szablonw klasy variant
//      "boost/varian/apply_visitor.hpp" - wizitacja typw wariantowych
//      "boost/varian/get.hpp"           - szablon funkcji get
//      "boost/varian/variant_fwd.hpp"   - definicja szablonu klasy visitor

// Potrzebne do napisania przykadu
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>

// Wizytator jest to klasa funkcyjna, ktra pozwala w bezpieczny sposb, ze wzgldu
// na typ, dokona pewnej operacji na typie wariantowym w zalenoci od tego jaki
// typ przechowuje
// Przykadowy wizytator wypisuje na standardowe wyjcie nazw typu i jego warto

class Wizytator : public boost::static_visitor<void>

//    Wizytator musi dziedziczy po klasie boost::static_visitor, ktra udostpnia
//    definicj typu (result_type).
{
public:

//     zaleca si aby parametrami przecionych operatorw () wizytatora byy referencje
//     poniewa wtedy unikniemy kosztownego kopiowania obiektw, a jeeli rozszerzymy zakres
//     typw zmiennej wariantowej nie bdzie uruchamiana nieodpowiednia metoda przez niejawn 
//     konwersj typw poniewa kod nie da si skompilowa.
    void operator()(int& i) const
    {
        std::cout << "To int: " << i << std::endl;
    }

    void operator()(std::string& s) const
    {
        std::cout << "To std::string: " << s << std::endl;
    }

    void operator()(double& d) const
    {
        std::cout << "To double: " << d << std::endl;
    }
};

int main(int argc, char *argv[])
{
//  deklaracja typu wariantowego, w tym wypadku przechowujcego int, std::string lub double
//  typy mog by zarwno typami wbudowanymi, jak rwnie dowoln klas jeeli przy
//  deklaracji nie poda si adnego parametru, zmienna wariantowa bdzie przechowywa
//  warto pierwszego z wymienionych typw, dlatego musi on dawa si skonstruowa
//  w sposb domylny. W tym przypadku przechowuje typ int
    boost::variant<int, std::string, double> example_variant;

    Wizytator wizytator;
    std::cout << "Wynik inicjalizacji bezparametrowej dla\nboost::variant<int, std::string, double>" << std::endl << std::endl;

//  funkcja boost::apply_visitor(wizytator, zmienna_wariantowa) pozwala na wizytacj
//  zmiennej za pomoc okrelonego wizytatora
    boost::apply_visitor(wizytator, example_variant);

//  boost::variant posiada rwnie konstruktor kopiujcy, ktry tworzy egzemplarz zmiennej wariantowej
//  z kopi parametru, typ parametru musi dawa si jednoznacznie skonwertowa na jeden z typw z zestawu
//  zmiennej wariantowej
    boost::variant<int, std::string, double> variant2("Napis");
    std::cout << std::endl << "Wynik inicjalizacji boost::variant<int, std::string, double>\nz parametrem \"Napis\"" << std::endl << std::endl;
    boost::apply_visitor(wizytator, variant2);

//  konstruktor pozwalajcy utworzy egzemplarz variant na podstawie innego egzemplarza, taka operacja jest
//  moliwa, jeeli kady typ z zestawu typw z parametru daje si skonwertowa na ktry z typw konstruowanych
    boost::variant<int, std::string, double> variant4(variant2);
    std::cout << std::endl << "Wynik inicjalizacji boost::variant<int, std::string, double>\nna podstawie innego egzemplarza" << std::endl << std::endl;
    boost::apply_visitor(wizytator, variant4);

//  poprawnie napisany wizytator nie obsuy typu, ktrego nie ma jawnie opisanego
//  zabezpiecza to przed konwersj niejawn
//  aby sprawdzi, e rzeczywicie si nie skompiluje naley odkomentowa ponisz lini
    boost::variant<int, std::string, double, char> variant3('a'); // zmienna wariantowa poszeona o char
//    boost::apply_visitor(wizytator, variant3);

//  destruktor usuwa obiekt i wywouje destruktor na rzecz przechowywanej wartoci, jeeli przechowuje wskanik nic nie robi
    variant3.~variant();
    variant4.~variant();

//  operator przypisania zamienia biec warto now. Typ parametru musi dawa si jednoznacznie skonwertowa
//  na jeden z typw z zestawu zmiennej wariantowej. Jeeli przechowywany typ jest identyczny z typem parametru
//  uywany jest kopiujcy operator przypisania, w innym przypadku uywany jest konstruktor kopiujcy, a bieca
//  warto jest porzucana
    example_variant = 1.1;
    std::cout << std::endl << "Wynik operatora przypisania dla parametru 1.1" << std::endl << std::endl;
    boost::apply_visitor(wizytator, example_variant);
    std::cout << std::endl << "Wynik operatora przypisania dla kolejno \"Napis\", 2, 'a', 2.0" << std::endl << std::endl;
    example_variant = "Napis";
    boost::apply_visitor(wizytator, example_variant);
    example_variant = 2;
    boost::apply_visitor(wizytator, example_variant);
    example_variant = 'a'; // tu char zosta niejawnie skonwertowany do int ('a' = 97)
    boost::apply_visitor(wizytator, example_variant);
    example_variant = 2.0;
    boost::apply_visitor(wizytator, example_variant);


//  metoda which zwraca indeks (od zera) typu, ktrego warto jest aktualnie przechowywana, wedug kolejnoci
//  zadeklarowanych typw
    std::cout << std::endl << "Wynik metody which() dla boost::variant<int, std::string, double>\nprzechowujcej warto typu double" << std::endl;
    std::cout << example_variant.which() << std::endl << std::endl;
    
//  metoda empty() zawsze zwraca false, jest wprowadzona tylko w celu zgodnoci z boost::any
    std::cout << "Wynik metody empty()" << std::endl;
    std::cout << example_variant.empty() << std::endl << std::endl;
    
//  metoda type() zwraca warto type_info() dla przechowywanej aktualnie wartoci
    std::cout << "Wynik metody type().name()" << std::endl;
    std::cout << example_variant.type().name() << std::endl << std::endl;

//  operator porwnania zwraca true jedynie gdy zgodne s typy i wartoci przechowywane w obu zmiennych wariantowych
    std::cout << "Wynik operatora = na dwch zmiennych wariantowych przechowujcych\n - rne typy\t\t\t\t";
    std::cout << (example_variant == variant2) << std::endl;
    example_variant = 2;
    variant2 = 3;
    std::cout << " - takie same typy, ale rne waroci\t";
    std::cout << (example_variant == variant2) << std::endl;
    variant2 = 2;
    std::cout << " - takie same typy i waroci\t\t";
    std::cout << (example_variant == variant2) << std::endl << std::endl;

//  analogicznie dziaa operator <, jeeli typu wartoci s rne zwraca which() < which(), dla obu zmiennych
//  jeeli typy s identyczne zwraca wynik operatora < dla wartoci.
    std::cout << "Wynik operatora < dla zmiennych wariantowych\nboost::variant<int, std::string, double> przechowujcych\n - rne typy (pierwsza int, druga double)\t";
    example_variant = 2;
    variant2 = 3.0;
    std::cout << (example_variant < variant2) << std::endl;
    std::cout << " - rne typy (pierwsza double, druga int)\t";
    example_variant = 2.0;
    variant2 = 3;
    std::cout << (example_variant < variant2) << std::endl;
    std::cout << " - takie same typy, (pierwsza = 2, druga = 3)\t";
    example_variant = 2;
    variant2 = 3;
    std::cout << (example_variant < variant2) << std::endl;
    std::cout << " - takie same typy, (pierwsza = 3, druga = 2)\t";
    example_variant = 3;
    variant2 = 2;
    std::cout << (example_variant < variant2) << std::endl << std::endl;

//  w bardzo prostych przypadkach mona zastosowa metod get,
//  metoda get<typename T>(boost::variant) zwraca warto typu T, jeeli zmienna wariantowa
//  przechowuje warto typu T, w przeciwnym razie zrzuca wyjtek boost::bad_get
//  testowanie wszystkich moliwoci t metod jest nieeleganckie i nieefektywne, w takim
//  celu stosuje si wizytatora
    std::cout << "Metoda get" << std::endl;
    example_variant = "Napis";
    try
    {
        double d = boost::get<double>(example_variant);
    }
    catch (boost::bad_get g)
    {
        std::cout << "Niezgodno typw (wyjtek)" << std::endl;
    }
    
//  jeeli nie chcemy obsugiwa wyjtku mona wywoa get ze wskanikiem na zmienn wariantow
//  wtedy przy niezgodnoci typw zwrcony bdzie wskanik pusty
    double* d = boost::get<double>(&example_variant);
    if (d == NULL)
       std::cout << "Niezgodno typw (wskanik)" << std::endl;

//  boost::variant pozawala m.in. na przechowywanie w jednym kontenerze heterogenicznych typw
    std::vector<boost::variant<int, std::string> > kontener;

//  pozwala to dodawa do kontenera elementy bez dbania o ich typ, o ile nale do zestawu zmiennej wariantowej
//  lub daj si jednoznacznie skonwertowa na jeden z typw w zestawie
    kontener.push_back(2);
    kontener.push_back("Napis");
    kontener.push_back(1);
    kontener.push_back("Inny Napis");
    
//  dziki odpowiedniemu zdefiniowaniu operatora < sortowanie ustawi wartoci tego samego typu obok siebie
//  i oczywicie posortowane wedug wartoci
    std::sort(kontener.begin(), kontener.end());
    
//  za pomoc wizytatora mona bez dbania o typ wartoci w kontenerze wykonywa operacj na elementach kontenera
//  np. wypisa na standardowe wyjcie
    std::cout << std::endl << "Wypisywanie zawartoci kontenera" << std::endl;
    
    std::vector<boost::variant<int, std::string> >::iterator it;
    for (it = kontener.begin(); it != kontener.end(); ++it)
        boost::apply_visitor(wizytator, *it);
    
    system("PAUSE");
    return 0;
}
